#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "renderer.h"
#include "device-output.h"

#undef OSDEBUG

#define UCLIMIT(x) ((t = (x)), (t &= ~(t>>31)), (t | ~((t-256) >> 31)))

#define RIDX 0
#define GIDX 1
#define BIDX 2    
#define TEX_WIDTH 512
#define TEX_HEIGHT 512
#define TEX_OFFSET(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3)
#define TEX_OFFSET_R(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3 + RIDX)
#define TEX_OFFSET_G(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3 + GIDX)
#define TEX_OFFSET_B(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3 + BIDX)

#include "FLImagePool.h"
#include "FLSharedMemory.h"

class ShmemOutputDevice : public OutputDevice
{
public:
  ShmemOutputDevice(const char *name);
  virtual int command(int argc, const char*const* argv);
  void *Address(void) { return cshm_->GetAddress(); }
  int isAttached(void) { return attached; }
private:
  void attach_shmem(int port);
  int attached;
  FLSharedMemory *cshm_;
};

class ShmemOutboard : public Renderer
{
public:
  ShmemOutboard(ShmemOutputDevice &device, int format);
  ~ShmemOutboard();
  virtual int command(int argc, const char*const* argv);
  virtual void setcolor(int c) {}

protected:
  void resize(int width, int height);
  ShmemOutputDevice &device_;
  FLImagePool *imagePool_;
  int scale_;
  int lastserial_;
  int lwidth_;
};

class ShmemOutboard_rgb : public ShmemOutboard
{
public:
  ShmemOutboard_rgb(ShmemOutputDevice &device);

protected:
  virtual int consume(const VideoFrame*);
};

class ShmemOutboard_yuv : public ShmemOutboard
{
public:
  ShmemOutboard_yuv(ShmemOutputDevice &device, int format)
    : ShmemOutboard(device, format) {};

protected:
  virtual int consume(const VideoFrame*);
  virtual void yuv_to_rgb(unsigned char *frame,
			  unsigned long ts) = 0;
};

class ShmemOutboard_422 : public ShmemOutboard_yuv
{
public:
  ShmemOutboard_422(ShmemOutputDevice &device);

protected:
  virtual void yuv_to_rgb(unsigned char *frame,
			  unsigned long ts);
};

class ShmemOutboard_411 : public ShmemOutboard_yuv
{
public:
  ShmemOutboard_411(ShmemOutputDevice &device);

protected:
  virtual void yuv_to_rgb(unsigned char *frame, unsigned long ts);
};

static ShmemOutputDevice shmem_device("shmem");

ShmemOutputDevice::ShmemOutputDevice(const char *name) :
  OutputDevice(name)
{

}

void ShmemOutputDevice::attach_shmem(int id)
{
#ifdef OSDEBUG
  fprintf(stderr, "about to construct new FLSharedMemory...");
#endif

  cshm_ = new FLSharedMemory(id);
  attached = 1;

#ifdef OSDEBUG
  fprintf(stderr, "done.\n");
#endif
}

int ShmemOutputDevice::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();
#ifdef OSDEBUG
  printf("ShmemOutputDevice::command argc=%d ", argc);
  for (int i = 0; i < argc; i++)
    printf("\"%s\" ", argv[i]);
  printf("\n");
#endif /* OSDEBUG */

  if (argc == 3) {
    if (strcmp(argv[1], "assistor") == 0)
      {
	if (strcmp(argv[2], "rgb") == 0)
	  {
	    printf("Assist with rgb\n");
	    ShmemOutboard *p = new ShmemOutboard_rgb(*this);
	    tcl.result(p->name());
	    return (TCL_OK);
	  }
      }

    if (strcmp(argv[1], "renderer") == 0) {
      if (strcmp(argv[2], "422") == 0) {

	ShmemOutboard* p = new ShmemOutboard_422(*this);
	tcl.result(p->name());
	return (TCL_OK);
      }

      if (strcmp(argv[2], "411") == 0) {
	ShmemOutboard* p = new ShmemOutboard_411(*this);
	tcl.result(p->name());
	return (TCL_OK);
      }
    }
    if (strcmp(argv[1], "id") == 0)
      {
#ifdef OSDEBUG
	fprintf(stderr, "Calling attach_shmem with %d\n", atoi(argv[2]));
#endif
	attach_shmem(atoi(argv[2]));
	return(TCL_OK);
      }
  }

  return (OutputDevice::command(argc, argv));
}

ShmemOutboard::ShmemOutboard(ShmemOutputDevice &device, int format) :
  Renderer(format), device_(device),
  imagePool_(0)
{
  printf("Construct ShmemOutboard for format %d\n", format);
    
  if(device_.isAttached())
    {
      imagePool_ = (FLImagePool *)device_.Address();
      if(imagePool_ != NULL) 
	{
	  imagePool_->_nextImage = -1;
	  imagePool_->_currentSerial = 0;
	  imagePool_->_clientActive = 1;
	}
    }
} 

ShmemOutboard::~ShmemOutboard()
{
  printf("Destruct ShmemOutboard()\n");
}

int ShmemOutboard::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();
#ifdef OSDEBUG
  printf("ShmemOutboard::command argc=%d ", argc);
  for (int i = 0; i < argc; i++)
    printf("\"%s\" ", argv[i]);
  printf("\n");
#endif /* OSDEBUG */

  if (argc == 3) {
    if (strcmp(argv[1], "scale") == 0) {
      int s = atoi(argv[2]);
      if (scale_ != s) {
	scale_ = s;
	resize(width_, height_);
      }
      return (TCL_OK);
    }
  }
  return (Renderer::command(argc, argv));
}

int ShmemOutboard_yuv::consume(const VideoFrame* frame)
{
  if (!samesize(frame))
    resize(frame->width_, frame->height_);

#ifdef OSDEBUG
  printf("Consume yuv %dx%d ts=%d\n", frame->width_, frame->height_,
	 frame->ts_);
#endif

  yuv_to_rgb(frame->bp_, frame->ts_);
	       
  return 0;
}

void ShmemOutboard::resize(int w, int h)
{
  width_ = w;
  height_ = h;
  framesize_ = w * h;

  if (width_ < TEX_WIDTH)
    {
      lwidth_ = width_;
    }
  else
    {
      lwidth_ = TEX_WIDTH;
    }
}

ShmemOutboard_rgb::ShmemOutboard_rgb(ShmemOutputDevice &device) :
  ShmemOutboard(device, FT_RGB)
{
  printf("Construct ShmemOutboard_rgb\n");
}

int ShmemOutboard_rgb::consume(const VideoFrame* frame)
{
  static int framect = 0;
  FLImageDescriptor *image;
  int nextIndex;

  RGBFrame *rgb = (RGBFrame *) frame;

  nextIndex = (imagePool_->_nextImage + 1) % imagePool_->_nImages;
  image = &(imagePool_->_images[nextIndex]);
  image->SetSize(rgb->width_, rgb->height_, 3);

  printf("Consume rgb %dx%d ts=%d len=%d\n", rgb->width_, rgb->height_,
	 rgb->ts_, rgb->len_);

  memcpy((void *)image->GetData(), rgb->bp_, rgb->len_);
  image->SetTimestamp(rgb->ts_);
	  
  return 1;
}

ShmemOutboard_422::ShmemOutboard_422(ShmemOutputDevice &device) :
  ShmemOutboard_yuv(device, FT_YUV_422)
{
  printf("Construct ShmemOutboard_422\n");
}

void ShmemOutboard_422::yuv_to_rgb(unsigned char *frame, unsigned long ts)
{
  unsigned char *yp, *up, *vp;
  unsigned char *yl, *ul, *vl;
  int w, h, t;
  int framesize;
  int y, u, v;
  unsigned char r, g, b;
  static int once = 0;

  FLImageDescriptor *image;
  volatile unsigned char *ImageData;
  int nextIndex;

  nextIndex = (imagePool_->_nextImage + 1) % imagePool_->_nImages;
  image = &(imagePool_->_images[nextIndex]);
  image->SetSize(width_, height_, 3);
  //  image->SetSize(TEX_WIDTH, TEX_HEIGHT, 3);
  ImageData = image->GetData();

#ifdef OSDEBUG      
      fprintf(stderr, "Index: %d\n", nextIndex);
#endif

  yp = frame;
  up = frame + framesize_;
  vp = up + framesize_ / 2;

  yl = yp;
  ul = up;
  vl = vp;

  for (h = 0; h < height_; h++)
    {
      yl = yp + h * width_;
      ul = up + h * width_ / 2;
      vl = vp + h * width_ / 2;
      for (w = 0; w < lwidth_; w += 2)
	{
	  u = (int) (*ul++ - 128);
	  v = (int) (*vl++ - 128);
	  
	  y = (int) (*yl++);

	  b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	  g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	  r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	  ImageData[TEX_OFFSET_R(w, h)] = r;
	  ImageData[TEX_OFFSET_G(w, h)] = g;
	  ImageData[TEX_OFFSET_B(w, h)] = b;

	  y = (int) (*yl++);

	  b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	  g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	  r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	  ImageData[TEX_OFFSET_R(w + 1, h)] = r;
	  ImageData[TEX_OFFSET_G(w + 1, h)] = g;
	  ImageData[TEX_OFFSET_B(w + 1, h)] = b;
	}
    }
  image->SetTimestamp(ts);
  imagePool_->_nextImage = nextIndex;
  imagePool_->_currentSerial++;
#ifdef OSDEBUG
  fprintf(stderr, "New frame ts=%d %dx%d\n", image->GetTimestamp(),
	  image->GetWidth(), image->GetHeight());
#endif
}

ShmemOutboard_411::ShmemOutboard_411(ShmemOutputDevice &device) :
  ShmemOutboard_yuv(device, FT_YUV_411)
{
  printf("Construct ShmemOutboard_411\n");
}

void ShmemOutboard_411::yuv_to_rgb(unsigned char *frame, unsigned long ts)
{
  unsigned char *yp, *up, *vp;
  int x, w, h, k, t;
  int framesize;
  double y, u, v;
  unsigned char r, g, b;
  static int once = 0;
  unsigned char *yl, *ul, *vl;

  FLImageDescriptor *image;
  volatile unsigned char *ImageData;
  int nextIndex;

  if(device_.isAttached())
    {
      nextIndex = (imagePool_->_nextImage + 1) % imagePool_->_nImages;
      image = &(imagePool_->_images[nextIndex]);
      ImageData = image->GetData();

#ifdef OSDEBUG      
      fprintf(stderr, "Index: %d\n", nextIndex);
#endif

      yp = frame;
      up = frame + framesize_;
      vp = up + framesize_ / 4;

      image->SetSize(width_, height_, 3);
      //  image->SetSize(TEX_WIDTH, TEX_HEIGHT, 3);

      //new
      for (h = 0; h < height_; h += 2)
	{
	  ul = up + (h / 2) * width_ / 2;
	  vl = vp + (h / 2) * width_ / 2;

	  // ul and vl are the uv for this and the next line

	  // First line
	  yl = yp + h * width_;
	
	  for (w = 0; w < lwidth_; w += 2)
	    {
	      u = (int) (*ul++ - 128);
	      v = (int) (*vl++ - 128);
	  
	      y = (int) (*yl++);

	      b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	      g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	      r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	      ImageData[TEX_OFFSET_R(w, h)] = r;
	      ImageData[TEX_OFFSET_G(w, h)] = g;
	      ImageData[TEX_OFFSET_B(w, h)] = b;

	      y = (int) (*yl++);

	      b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	      g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	      r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	      ImageData[TEX_OFFSET_R(w + 1, h)] = r;
	      ImageData[TEX_OFFSET_G(w + 1, h)] = g;
	      ImageData[TEX_OFFSET_B(w + 1, h)] = b;
	    }

	  // Next line; reset ul/vl and bump yl

	  ul = up + (h / 2) * width_ / 2;
	  vl = vp + (h / 2) * width_ / 2;
	  yl = yp + (h + 1) * width_;
	
	  for (w = 0; w < lwidth_; w += 2)
	    {
	      u = (int) (*ul++ - 128);
	      v = (int) (*vl++ - 128);
	  
	      y = (int) (*yl++);

	      b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	      g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	      r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	      ImageData[TEX_OFFSET_R(w, (h+1))] = r;
	      ImageData[TEX_OFFSET_G(w, (h+1))] = g;
	      ImageData[TEX_OFFSET_B(w, (h+1))] = b;

	      y = (int) (*yl++);

	      b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	      g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	      r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	      ImageData[TEX_OFFSET_R(w + 1, (h+1))] = r;
	      ImageData[TEX_OFFSET_G(w + 1, (h+1))] = g;
	      ImageData[TEX_OFFSET_B(w + 1, (h+1))] = b;
	    }
	}
    
#if 0
      for (h = 0; h < height_; h++)
	{
	  x = 0;
	  for (w = 0; w < width_; w += 2)
	    {
	      for (k = 0; k < 2; k++)
		{
		  r = g = b = (yp[yi++]);

		  if (w % 5 != 0)
		    {
		      if (x < TEX_WIDTH)
			{
			  image_->data[TEX_OFFSET_R(x, h)] = r;
			  image_->data[TEX_OFFSET_G(x, h)] = g;
			  image_->data[TEX_OFFSET_B(x, h)] = b;
			}
		      x++;
		    }
		}
	    }
	}
#endif
      image->SetTimestamp(ts);
      imagePool_->_nextImage = nextIndex;
      imagePool_->_currentSerial++;
#ifdef OSDEBUG
      fprintf(stderr,"New 411 frame ts=%d %dx%d\n", image->GetTimestamp(),
	      image->GetWidth(), image->GetHeight());
#endif
    }
}
